#include "stdafx.h"
#include "mm.h"
#include "feature_3d.h"
#include "diff_wheel_robot.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

vector<sFeature3D> _feature_3d;

bool load_feature3D (const char *file_name)
{
	_feature_3d.clear ();
	_feature_3d.reserve (1000);

	FILE *fp = fopen (file_name, "r");
	if (fp) {
		char word[1024+1], *p;
		while (fgets (word, 1024, fp)) {
			word[1024] = '\0';
			for (p = word; *p==' ' || *p=='\t'; ++p);

			if (*p == '#') continue;

			sFeature3D feature;

			feature.x = strtod (p, &p);		if (*p=='\r' || *p=='\n' || *p=='\0') continue;
			feature.y = strtod (p, &p);		if (*p=='\r' || *p=='\n' || *p=='\0') continue;
			feature.z = strtod (p, &p);		if (*p=='\r' || *p=='\n' || *p=='\0') continue;
			feature.r = strtod (p, &p);
			feature.active = false;
			
			_feature_3d.push_back (feature);
		}	
		fclose (fp);

		return true;
	}
	return false;
}


void calc_seen_feature3D ()
{
	const double detection_range_min = 1.;
	const double detection_range_max = 3.;
	const double detection_angle = M_PI/2;

	for (unsigned int i=0; i<_feature_3d.size(); ++i) {
		double dx = _feature_3d[i].x - _robot_x;
		double dy = _feature_3d[i].y - _robot_y;
		double dz = _feature_3d[i].z - 0.;

		double dl = sqrt (dx*dx + dy*dy + dz*dz);
		double dth = DeltaRad (atan2 (dy, dx), _robot_theta);
		double dal = acos (sqrt (dx*dx + dy*dy)/dl);

		if ((detection_range_min <= dl) && (dl <= detection_range_max) && 
			(-detection_angle/2. < dth && dth < detection_angle/2.)) {
			_feature_3d[i].active = true;
		}
		else {
			_feature_3d[i].active = false;
		}
	}
}
