#include "stdafx.h"
#include <math.h>
#include "floor_landmark.h"
#include "diff_wheel_robot.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

map<unsigned long long, sFloorLandmark> _floor_landmark;

void calc_floor_landmark ()
{
	for (map<unsigned long long, sFloorLandmark>::iterator it = _floor_landmark.begin(); it != _floor_landmark.end(); it++) {
		sFloorLandmark &feature = it->second;

		double dx = feature.x - _robot_x;
		double dy = feature.y - _robot_y;
		double dist = sqrt(dx*dx + dy*dy);

		feature.active = (dist <= feature.r) ? true : false;
	}
}

sFloorLandmark *get_floor_landmark (unsigned long long uid)
{
	map<unsigned long long, sFloorLandmark>::iterator it = _floor_landmark.find (uid);
	if (it != _floor_landmark.end()) {
		return &it->second;
	}

	return NULL;
}

bool load_floor_landmark(const char *file_name)
{
	_floor_landmark.clear ();

	FILE *fp = fopen (file_name, "r");
	if (fp) {
		char word[1024+1], *p;
		while (fgets (word, 1024, fp)) {
			word[1024] = '\0';
			p = word;

			if (*p == '#') continue;

			int no    = strtol (p, &p, 10);
			double sx = strtod (p, &p);
			double sy = strtod (p, &p);
			double ex = strtod (p, &p);
			double ey = strtod (p, &p);
			double r  = strtod (p, &p);

			for (int i=0; i<no; ++i) {
				unsigned long long uid = _strtoui64 (p, &p, 16);
				if (uid) {
					sFloorLandmark info;
					info.uid = uid;
					info.x   = sx + (double)i/(no - 1)*(ex - sx);
					info.y   = sy + (double)i/(no - 1)*(ey - sy);
					info.r   = r;
	
					_floor_landmark.insert (make_pair<unsigned long long, sFloorLandmark>(uid, info));
				}
			}
		}	
		fclose (fp);

		return true;
	}
	return false;
}

