
// mobile_robot_simulationDlg.cpp : implementation file
//

#include "stdafx.h"
#include "mobile_robot_simulation.h"
#include "mobile_robot_simulationDlg.h"
#include "MultimediaTimer.h"
#include "mm.h"
#include "AppIni.h"
#include <math.h>

#include "map_2d.h"
#include "diff_wheel_robot.h"
#include "feature_2d.h"
#include "feature_3d.h"
#include "ceiling_landmark.h"
#include "floor_landmark.h"
#include "scanning_laser.h"
#include "beacon.h"

CMultimediaTimer mt;

extern void RegisteryClearRestrictRemoteClients ();
extern UINT __cdecl rpc_server(LPVOID pParam);

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


// Cmobile_robot_simulationDlg dialog
Cmobile_robot_simulationDlg::Cmobile_robot_simulationDlg(CWnd* pParent /*=NULL*/)
	: CDialog(Cmobile_robot_simulationDlg::IDD, pParent)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);

	_sel = SEL_NONE;

	_cursorDragStartPt.x = 0;
	_cursorDragStartPt.y = 0;
}

void Cmobile_robot_simulationDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
}

BEGIN_MESSAGE_MAP(Cmobile_robot_simulationDlg, CDialog)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	//}}AFX_MSG_MAP
	ON_WM_TIMER()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_MOUSEWHEEL()
	ON_WM_LBUTTONDBLCLK()
	ON_BN_CLICKED(IDC_BUTTON_LOAD_MAP, &Cmobile_robot_simulationDlg::OnBnClickedButtonLoadMap)
	ON_WM_DESTROY()
END_MESSAGE_MAP()


int Cmobile_robot_simulationDlg::M2Px(double _x) { return INTEGER( _scale*(_x - _offset_x)); }
int Cmobile_robot_simulationDlg::M2Py(double _y) { return INTEGER(-_scale*(_y - _offset_y)) + _mapRect.Height(); }
int Cmobile_robot_simulationDlg::M2Pl(double _d) { return INTEGER( _scale*_d); }
double Cmobile_robot_simulationDlg::P2Mx(int _x) { return (                    _x)/_scale + _offset_x; }
double Cmobile_robot_simulationDlg::P2My(int _y) { return (_mapRect.Height() - _y)/_scale + _offset_y; }
double Cmobile_robot_simulationDlg::P2Ml(int _d) { return (                    _d)/_scale; }

// Cmobile_robot_simulationDlg message handlers
BOOL Cmobile_robot_simulationDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	//  ׸   ũ⸦ ´.
	GetDlgItem (IDC_STATIC_MAP)->GetWindowRect (_mapRect);
	GetDlgItem (IDC_STATIC_MAP)->DestroyWindow();
	ScreenToClient(_mapRect);

	// TODO: Add extra initialization here
	_robot_x    = g_ini.robot_x;
	_robot_y    = g_ini.robot_y;
	_robot_theta= g_ini.robot_theta;

	_scale    = g_ini.scale;
	_offset_x = g_ini.offset_x;
	_offset_y = g_ini.offset_y;

	load_dxf (g_ini.map_file, &_map_2d);
	load_beacon (".\\map\\beacon.txt");
	load_feature2D (".\\map\\feature_2d.txt");
	load_feature3D (".\\map\\feature_3d.txt");
	load_floor_landmark (".\\map\\floor_landmark.txt");
	load_ceiling_landmark (".\\map\\ceiling_landmark.txt");

	// load_dxf() Լ о   mm Ǿִ. 
	//  ͸ m  Ѵ.
	for (unsigned int i=0; i<_map_2d.size(); ++i) {
		_map_2d[i].x0 *= 0.001;
		_map_2d[i].y0 *= 0.001;
		_map_2d[i].z0 *= 0.001;
		_map_2d[i].x1 *= 0.001;
		_map_2d[i].y1 *= 0.001;
		_map_2d[i].z1 *= 0.001; 
	}

	RegisteryClearRestrictRemoteClients ();

	AfxBeginThread (rpc_server, THREAD_PRIORITY_NORMAL);

	QueryPerformanceFrequency (&_freq);
	QueryPerformanceCounter (&_time_prev);

	CheckDlgButton (IDC_CHECK_MAP				, g_ini.show_map			? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton (IDC_CHECK_2D_FEATURE		, g_ini.show_2d_feature		? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton (IDC_CHECK_3D_FEATURE		, g_ini.show_3d_feature		? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton (IDC_CHECK_BEACON			, g_ini.show_beacon			? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton (IDC_CHECK_CEILING_LANDMARK	, g_ini.show_ceiling_landmark ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton (IDC_CHECK_FLOOR_LANDMARK	, g_ini.show_floor_landmark ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton (IDC_CHECK_SCANNING_LASER	, g_ini.show_scanning_laser ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton (IDC_CHECK_ROBOT				, g_ini.show_robot			? BST_CHECKED : BST_UNCHECKED);

	SetTimer (1000, 10, NULL);
	mt.Start (1);

	return TRUE;  // return TRUE  unless you set the focus to a control
}


void Cmobile_robot_simulationDlg::OnDestroy()
{
	CDialog::OnDestroy();

	// TODO: Add your message handler code here
	KillTimer (1000);
	mt.Stop ();

	g_ini.robot_x = _robot_x;
	g_ini.robot_y = _robot_y;
	g_ini.robot_theta = _robot_theta;

	g_ini.scale	   = _scale;
	g_ini.offset_x = _offset_x;
	g_ini.offset_y = _offset_y;

	g_ini.show_map				= IsDlgButtonChecked (IDC_CHECK_MAP				);
	g_ini.show_2d_feature		= IsDlgButtonChecked (IDC_CHECK_2D_FEATURE		);
	g_ini.show_3d_feature		= IsDlgButtonChecked (IDC_CHECK_3D_FEATURE		);
	g_ini.show_beacon			= IsDlgButtonChecked (IDC_CHECK_BEACON			);
	g_ini.show_ceiling_landmark	= IsDlgButtonChecked (IDC_CHECK_CEILING_LANDMARK);
	g_ini.show_floor_landmark	= IsDlgButtonChecked (IDC_CHECK_FLOOR_LANDMARK	);
	g_ini.show_scanning_laser	= IsDlgButtonChecked (IDC_CHECK_SCANNING_LASER	);
	g_ini.show_robot			= IsDlgButtonChecked (IDC_CHECK_ROBOT			);
}

void Cmobile_robot_simulationDlg::OnTimer(UINT_PTR nIDEvent)
{
	LARGE_INTEGER tc;
	QueryPerformanceCounter(&tc);
	double dt = (double)(tc.QuadPart - _time_prev.QuadPart)/_freq.QuadPart;
	_time_prev = tc;

	robot_move (dt);
	calc_seen_beacon ();
	calc_seen_feature2D ();
	calc_seen_feature3D ();
	calc_laser_value ();
	calc_floor_landmark ();
	calc_seen_ceiling_landmark ();

	if (_memDC.GetSafeHdc()) {
		DrawBackground (_memDC, 100, 100);

		if (IsDlgButtonChecked(IDC_CHECK_MAP				) == BST_CHECKED) DrawMap (_memDC);
		if (IsDlgButtonChecked(IDC_CHECK_2D_FEATURE			) == BST_CHECKED) DrawFeature2D (_memDC);
		if (IsDlgButtonChecked(IDC_CHECK_3D_FEATURE			) == BST_CHECKED) DrawFeature3D (_memDC);
		if (IsDlgButtonChecked(IDC_CHECK_BEACON				) == BST_CHECKED) DrawBeacon (_memDC);
		if (IsDlgButtonChecked(IDC_CHECK_CEILING_LANDMARK	) == BST_CHECKED) DrawCeilingLandmark (_memDC);
		if (IsDlgButtonChecked(IDC_CHECK_FLOOR_LANDMARK		) == BST_CHECKED) DrawFloorLandmark (_memDC);
		if (IsDlgButtonChecked(IDC_CHECK_SCANNING_LASER		) == BST_CHECKED) DrawScanningLaser (_memDC, RGB(255,0,0), false);
		if (IsDlgButtonChecked(IDC_CHECK_ROBOT				) == BST_CHECKED) DrawRobot (_memDC, _robot_x, _robot_y, _robot_theta, _axle_length, RGB(0,0,255));

		DrawText (_memDC, dt);
	}

	InvalidateRect (&_mapRect, FALSE);

	CDialog::OnTimer(nIDEvent);
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void Cmobile_robot_simulationDlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CPaintDC dc(this); // device context for painting

		if (_memDC.GetSafeHdc() == NULL) {
			_memDC.CreateCompatibleDC(&dc);
			_bitmap.CreateCompatibleBitmap(&dc, _mapRect.Width(), _mapRect.Height());
			_bitmapOld = _memDC.SelectObject(&_bitmap);
			ASSERT (_bitmapOld);
		}

		if (_memDC.GetSafeHdc() != NULL) {
			dc.BitBlt(_mapRect.left, _mapRect.top, _mapRect.Width(), _mapRect.Height(), &_memDC, 0,0, SRCCOPY);
		}

		CDialog::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR Cmobile_robot_simulationDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

void Cmobile_robot_simulationDlg::DrawText (CDC &dc, double dt)
{
	char text[1024];
	sprintf (text, "delta time = %.1f (ms)", dt*1000.);
	dc.TextOut (20, 10, text, strlen(text));

	sprintf (text, "robot pos(x,y,theta) = %.3f, %.3f, %.1f (m,m,deg)", _robot_x, _robot_y, _RAD2DEG*_robot_theta);
	dc.TextOut (20, 30, text, strlen(text));

	sprintf (text, "Wheel ang(left,right) = %.1f, %.1f (deg,deg)", _RAD2DEG*_current_position_left, _RAD2DEG*_current_position_right);
	dc.TextOut (20, 50, text, strlen(text));
}

void Cmobile_robot_simulationDlg::DrawBackground (CDC &dc, int width, int height)
{
	dc.SelectStockObject (WHITE_PEN);
	dc.SelectStockObject (WHITE_BRUSH);

	dc.Rectangle (0, 0, 10000, 10000);

	CPen pen (PS_SOLID, 1, RGB(225,225,225));
	dc.SelectObject (pen);

	for (int i=0; i<=width; ++i) {
		dc.MoveTo (M2Px(i), M2Py(0));
		dc.LineTo (M2Px(i), M2Py(height));
	}
	for (int i=0; i<=height; ++i) {
		dc.MoveTo (M2Px(0), M2Py(i));
		dc.LineTo (M2Px(width), M2Py(i));
	}
}

void Cmobile_robot_simulationDlg::DrawMap (CDC &dc)
{
	CPen pen (PS_SOLID, 1, RGB(127,127,127));
	dc.SelectObject (pen);

	for (unsigned int i=0; i<_map_2d.size(); ++i) {
		if (_map_2d[i].type == ET_LINE) {
			dc.MoveTo (M2Px(_map_2d[i].x0), M2Py(_map_2d[i].y0));
			dc.LineTo (M2Px(_map_2d[i].x1), M2Py(_map_2d[i].y1));
		}
	}
}

void Cmobile_robot_simulationDlg::DrawRobot (CDC &dc, double x, double y, double theta, double diameter, COLORREF color)
{
	dc.SelectStockObject (NULL_BRUSH);

	CPen pen (PS_SOLID, 2, color);
	dc.SelectObject (pen);

	int rx = M2Px(x);
	int ry = M2Py(y);
	int rr = M2Pl(diameter/2);

	int dx = (int)(rr*cos(theta));
	int dy = (int)(rr*sin(theta));

	dc.Ellipse (rx-rr, ry-rr, rx+rr, ry+rr);
	dc.MoveTo (rx, ry);
	dc.LineTo (rx+dx, ry-dy);
}

void Cmobile_robot_simulationDlg::DrawScanningLaser (CDC &dc, COLORREF color, bool line)
{
	CPen pen (PS_SOLID, 1, color);
	dc.SelectObject (pen);

	double dang = _laser_scan_range/(_laser_scan_count - 1);

	for (int i=0; i<_laser_scan_count; ++i) {
		double ang = _robot_theta + _laser_theta - _laser_scan_range/2 + i*dang;

		double x0 = _robot_x + _laser_x*cos(_robot_theta) - _laser_y*sin(_robot_theta);
		double y0 = _robot_y + _laser_x*sin(_robot_theta) - _laser_y*cos(_robot_theta);

		double x1 = x0 + _laser_value[i]*cos(ang);
		double y1 = y0 + _laser_value[i]*sin(ang);

		if (line) {
			dc.MoveTo (M2Px(x0), M2Py(y0));
			dc.LineTo (M2Px(x1), M2Py(y1));
		}
		else {
			int x = M2Px(x1);
			int y = M2Py(y1);
			dc.SetPixel (x,   y,   color);
			dc.SetPixel (x,   y+1, color);
			dc.SetPixel (x+1, y,   color);
			dc.SetPixel (x+1, y+1, color);
		}
	}
}

void Cmobile_robot_simulationDlg::DrawFeature2D (CDC &dc)
{
	CPen pen (PS_SOLID, 1, RGB(0,64,0));
	dc.SelectObject (pen);

	CBrush brush_active ( RGB(0,255,0));
	CBrush brush_inactive ( RGB(0,127,0));

	for (unsigned int i=0; i<_feature_2d.size(); ++i) {
		int x = M2Px(_feature_2d[i].x);
		int y = M2Py(_feature_2d[i].y);
		int r = M2Pl(_feature_2d[i].r/2);
		
		dc.SelectObject (_feature_2d[i].active ? brush_active : brush_inactive);
		dc.Ellipse (x-r, y-r, x+r, y+r);
	}
}

void Cmobile_robot_simulationDlg::DrawFeature3D (CDC &dc)
{
	CPen pen (PS_SOLID, 1, RGB(0,64,0));
	dc.SelectObject (pen);

	CBrush brush_active ( RGB(0,255,0));
	CBrush brush_inactive ( RGB(0,127,0));

	for (unsigned int i=0; i<_feature_3d.size(); ++i) {
		int x0 = M2Px(_feature_3d[i].x);
		int y0 = M2Py(_feature_3d[i].y);
		int x = M2Px(_feature_3d[i].x + _feature_3d[i].z/5.);
		int y = M2Py(_feature_3d[i].y + _feature_3d[i].z/5.);
		int r = M2Pl(_feature_3d[i].r/2);
		
		dc.MoveTo (x0, y0);
		dc.LineTo (x, y);

		dc.SelectObject (_feature_3d[i].active ? brush_active : brush_inactive);
		dc.Ellipse (x-r, y-r, x+r, y+r);
	}
}

void Cmobile_robot_simulationDlg::DrawFloorLandmark (CDC &dc)
{
	CPen pen_active (PS_SOLID, 1, RGB(255,64,64));
	CPen pen_inactive (PS_SOLID, 1, RGB(128,128,128));

	dc.SelectStockObject (NULL_BRUSH);

	for (map<unsigned long long, sFloorLandmark>::iterator it = _floor_landmark.begin(); it != _floor_landmark.end(); it++) {
		sFloorLandmark &landmark = it->second;

		int x = M2Px(landmark.x);
		int y = M2Py(landmark.y);
		int r = M2Pl(landmark.r);

		dc.SelectObject (landmark.active ? pen_active : pen_inactive);
		dc.Ellipse (x-r, y-r, x+r, y+r);
	}
}

void Cmobile_robot_simulationDlg::DrawBeacon (CDC &dc)
{
	CPen pen_active (PS_SOLID, 1, RGB(255,64,64));
	CPen pen_inactive (PS_SOLID, 1, RGB(128,128,128));

	dc.SelectStockObject (NULL_BRUSH);

	for (unsigned int i=0; i<_beacon.size(); ++i) {
		int x = M2Px(_beacon[i].x);
		int y = M2Py(_beacon[i].y);
		int r = M2Pl(_beacon[i].r);

		dc.SelectObject (_beacon[i].active ? pen_active : pen_inactive);
		dc.Ellipse (x-r, y-r, x+r, y+r);
	}
}

void Cmobile_robot_simulationDlg::DrawCeilingLandmark (CDC &dc)
{
	CPen pen_active (PS_SOLID, 1, RGB(255,64,64));
	CPen pen_inactive (PS_SOLID, 1, RGB(128,128,128));

	dc.SelectStockObject (NULL_BRUSH);

	for (unsigned int i=0; i<_ceiling_landmark.size(); ++i) {
		int x = M2Px(_ceiling_landmark[i].x);
		int y = M2Py(_ceiling_landmark[i].y);
		int r = M2Pl(_ceiling_landmark[i].r);

		dc.SelectObject (_ceiling_landmark[i].active ? pen_active : pen_inactive);
		dc.Ellipse (x-r, y-r, x+r, y+r);
	}
}

void Cmobile_robot_simulationDlg::OnLButtonDown(UINT nFlags, CPoint point)
{
	if (_mapRect.left < point.x && _mapRect.top < point.y && point.x < _mapRect.right && point.y < _mapRect.bottom ){
		// κ 콺 Ʈ Ÿ Ͽ κ ݰ溸  κ Ѵ.
		double dist = Distance (P2Mx(point.x - _mapRect.left), P2My(point.y - _mapRect.top), _robot_x, _robot_y);
		
		if (dist < _axle_length/2.) {
			_sel = SEL_CURRENT;
		}
		else {
			_sel = SEL_MAP;
		}

		SetCapture();
		SetFocus ();
		_cursorDragStartPt = point;
	}
	else {
		_sel = SEL_NONE;
	}

	CDialog::OnLButtonDown(nFlags, point);
}

void Cmobile_robot_simulationDlg::OnLButtonUp(UINT nFlags, CPoint point)
{
	if (_sel == SEL_CURRENT) {
	}
	else if (_sel == SEL_MAP) {
	}
	ReleaseCapture();

	CDialog::OnLButtonUp(nFlags, point);
}

void Cmobile_robot_simulationDlg::OnLButtonDblClk(UINT nFlags, CPoint point)
{
	_robot_x = P2Mx(point.x-_mapRect.left);
	_robot_y = P2My(point.y-_mapRect.top);

	_sel = SEL_CURRENT;

	CDialog::OnLButtonDblClk(nFlags, point);
}

void Cmobile_robot_simulationDlg::OnMouseMove(UINT nFlags, CPoint point)
{
	if (GetCapture()==this) {
		CPoint delta = point - _cursorDragStartPt;
		_cursorDragStartPt = point;

		if (_sel == SEL_CURRENT) {
			_robot_x += P2Ml(delta.x);
			_robot_y -= P2Ml(delta.y);
		}
		else if (_sel == SEL_MAP) {
			_offset_x -= P2Ml(delta.x);
			_offset_y += P2Ml(delta.y);
		}
	}

	CDialog::OnMouseMove(nFlags, point);
}

BOOL Cmobile_robot_simulationDlg::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt)
{
	if (_sel == SEL_CURRENT){
		_robot_theta += zDelta/1000.;
	}
	else if (_sel == SEL_MAP) {
		double newScale = bound (_scale*(1. + zDelta/1000.), 1., 1000.);
		double d = newScale/_scale;

		_scale = newScale;

		_offset_x += P2Ml(_mapRect.Width() )/2.*(1. - 1./d); 
		_offset_y += P2Ml(_mapRect.Height())/2.*(1. - 1./d);
	}

	return CDialog::OnMouseWheel(nFlags, zDelta, pt);
}

void Cmobile_robot_simulationDlg::OnBnClickedButtonLoadMap()
{
	CFileDialog fdlg(TRUE, "dxf", "*.dxf", OFN_FILEMUSTEXIST| OFN_HIDEREADONLY, 
		"Map Files (*.dxf)|*.dxf|All Files (*.*)|*.*||", this);

	fdlg.DoModal ();

	CString fileName = fdlg.GetPathName ();

	strncpy (g_ini.map_file, (LPCSTR)fileName, MAX_PATH);

	_map_2d.clear();
	load_dxf ((LPCSTR)fileName, &_map_2d);

	// load_dxf() Լ о   mm Ǿִ. 
	//  ͸ m  Ѵ.
	for (unsigned int i=0; i<_map_2d.size(); ++i) {
		_map_2d[i].x0 *= 0.001;
		_map_2d[i].y0 *= 0.001;
		_map_2d[i].z0 *= 0.001;
		_map_2d[i].x1 *= 0.001;
		_map_2d[i].y1 *= 0.001;
		_map_2d[i].z1 *= 0.001; 
	}
}

