#include "stdafx.h"
#include "mm.h"
#include "plane_sweep_algorithm.h"
#include "scanning_laser.h"
#include "diff_wheel_robot.h"
#include <vector>
#include <algorithm>

using namespace std;

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

sVertex *_vertex_list;
sEdge *_edge_list;

int _no_vertex;
int _no_edge;

double line_intersect(double x0, double y0, double x1, double y1, double x2, double y2, double x3, double y3)
{
	// (x0, y0), (x1, y1) մ а (x2, y2), (x3, y3) մ  ϴ ˻Ѵ.
	//  Ѵٸ (x0, y0)  ϴ  Ÿ ϰ,
	//  ʴ´ٸ  ĳ ִ    Ѵ.

	double fx = x3 - x2;
	double fy = y3 - y2;

	double ex = x1 - x0;
	double ey = y1 - y0;

	double t1 = ((x2*fy - y2*fx) - (x0*fy - y0*fx))/(ex*fy - ey*fx);
	double t2 = ((x0*ey - y0*ex) - (x2*ey - y2*ex))/(fx*ey - fy*ex);

	if (0 <= t1 && t1 <= 1. && 0. <= t2 && t2 <= 1.) {
		double dx = t1*ex;
		double dy = t1*ey;

		return sqrt(dx*dx + dy*dy);
	}
	else {
		return _laser_max_range;
	}
}

double is_visible (vector<sEdge *> &S, sVertex &v, sVertex &v_i)
{
	double min_d = _laser_max_range;

	for (unsigned int i=0; i<S.size(); ++i) {
		if (S[i]) {
			double d = line_intersect (v.x, v.y, v_i.x, v_i.y, S[i]->s->x, S[i]->s->y, S[i]->e->x, S[i]->e->y);
			if (d < min_d) {
				min_d = d;
			}
		}
	}
	return min_d;
}

bool in_list(vector<sEdge *> &S, sEdge *e)
{
	// Ʈ S ִ  Ȯ

	for(unsigned int i = 0; i<S.size(); i++) {
		if (S[i] && S[i] == e) {
			return true;
		}
	}
	return false;
}

bool Compare(const sVertex *elem1, const sVertex *elem2 )
{
	return elem1->alpha < elem2->alpha;
}

void plane_sweep_algorithm()
{
	// *** å Principles of Robot Motion, p.114   *** 

	//   
	double sensor_begin = DeltaRad(_robot_theta + _laser_theta - _laser_scan_range/2, 0.);

	sVertex v = { 
		_robot_x + _laser_x*cos(_robot_theta) - _laser_y*sin(_robot_theta), 
		_robot_y + _laser_x*sin(_robot_theta) - _laser_y*cos(_robot_theta), 
	};

	vector<sVertex *> E;	// vertex list
	vector<sEdge *> S;		// active list

	E.reserve (_no_vertex);
	S.reserve (_no_edge);

	// (vertex) v Ͽ ٸ      ̷ (alpha) Ѵ.
	// ׸ alpha Ͽ  ĵ vertex list E .
	for(int j = 0; j<_no_vertex; j++) {
		double alpha = atan2(_vertex_list[j].y - v.y, _vertex_list[j].x - v.x);
		if (alpha < sensor_begin) alpha += 2*M_PI;

		_vertex_list[j].alpha = alpha;

		E.push_back(&_vertex_list[j]);
	}
	sort(E.begin(), E.end(), Compare);

	// edge ߿,   alpha    alpha ū , ۰   ġ ٲ۴.
	for (int j = 0; j < _no_edge; j++) {
		if (_edge_list[j].e->alpha < _edge_list[j].s->alpha) {
			sVertex *t = _edge_list[j].e;
			_edge_list[j].e = _edge_list[j].s;
			_edge_list[j].s = t;
		}
	}

	// v   (half-line) ϴ edge Ʈ  S .
	for (int j=0; j < _no_edge; ++j) {
		double d = line_intersect(v.x, v.y, 
			v.x + _laser_max_range*cos(sensor_begin), v.y + _laser_max_range*sin(sensor_begin), 
			_edge_list[j].s->x, _edge_list[j].s->y, _edge_list[j].e->x, _edge_list[j].e->y);

		if (d < _laser_max_range) {
			S.push_back(&_edge_list[j]);
		}
	}

	unsigned int jj = 0;
	for(int i = 0; i<_laser_scan_count; i++) {
		double alpha_s = sensor_begin + i*_laser_scan_range/(_laser_scan_count - 1);

		sVertex v_i = { 
			v.x + _laser_max_range*cos(alpha_s),
			v.y + _laser_max_range*sin(alpha_s),
			};

		// v v_i ̸ edge(v, v_i) visibility graph ߰Ѵ.
		_laser_value[i] = is_visible (S, v, v_i);

		for ( ; jj < E.size() && E[jj]->alpha < alpha_s + _DEG2RAD*1.; jj++) {
			//  v_i   edge E Ʈ S , E S Ѵ.
			// insert_if_vertex_is_begin(S, v_i);
			for(int k = 0; k<_no_edge; k++) {
				// edge ̰ S   ƴ Ȯ
				if(_edge_list[k].s == E[jj] && !in_list(S, &_edge_list[k])) {
					S.push_back(&_edge_list[k]);
				}
			}
			
			//  v_i S ִ edge  ̸ Sκ Ѵ.
			for(unsigned int k = 0; k<S.size(); k++) {
				if (S[k] && S[k]->e == E[jj]) {
					// ¥  ʰ, NULLθ ǥصд.
					// S.erase(S.begin() + k); 
					S[k] = NULL;
				}
			}
		}
	}
}
