
#include "stdafx.h"

#include "mm.h"
#include "plane_sweep_algorithm.h"
#include "map_2d.h"
#include "diff_wheel_robot.h"
#include "scanning_laser.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


// κ ǥ踦   ġ ġ
double _laser_x = 0.25;
double _laser_y = 0.0;
double _laser_theta = _DEG2RAD*0.;
//  ּ, ִ  Ÿ
double _laser_min_range = 0.03;
double _laser_max_range = 8.;
//  ĵ      ػ
double _laser_scan_range = _DEG2RAD*180.;
//     ϴ 迭
int _laser_scan_count = 181;
double _laser_value[361] = {0., };


double map_intersection (double x0, double y0, double x1, double y1)
{
	//   ((x0, y0) - (x1, y1))    а  ˻Ѵ.
	// Ѵٸ, (x0, y0)   Ÿ Ѵ.
	//  ʴ´ٸ,  ɳ ִ   Ѵ.

	double min_l = _laser_max_range;
	
	for (unsigned int i=0; i<_map_2d.size(); ++i) {
		double cx = _map_2d[i].x0;
		double cy = _map_2d[i].y0;
		double fx = _map_2d[i].x1 - _map_2d[i].x0;
		double fy = _map_2d[i].y1 - _map_2d[i].y0;

		double ax = x0;
		double ay = y0;
		double ex = x1 - x0;
		double ey = y1 - y0;

		double t1 = ((cx*fy - cy*fx) - (ax*fy - ay*fx))/(ex*fy - ey*fx);
		double t2 = ((ax*ey - ay*ex) - (cx*ey - cy*ex))/(fx*ey - fy*ex);
	
		if (0. <= t1 && t1 <= 1. && 0. <= t2 && t2 <= 1.) {
			double dx = t1*ex;
			double dy = t1*ey;

			double l = sqrt(dx*dx + dy*dy);
			if (l < min_l) min_l = l;
		}
	}
	return min_l;
}

void calc_laser_value_ ()
{
	//       а  ˻Ѵ.
	//   360̰   n 360*n   ˻縦 ϰԵȴ.
	//  ȿ ˻ ̴.

	double dang = _laser_scan_range/(_laser_scan_count - 1);

	for (int i=0; i<_laser_scan_count; ++i) {
		double ang = _robot_theta + _laser_theta - _laser_scan_range/2 + i*dang;

		double x0 = _robot_x + _laser_x*cos(_robot_theta) - _laser_y*sin(_robot_theta);
		double y0 = _robot_y + _laser_x*sin(_robot_theta) - _laser_y*cos(_robot_theta);

		double x1 = x0 + _laser_max_range*cos(ang);
		double y1 = y0 + _laser_max_range*sin(ang);

		_laser_value[i] = map_intersection (x0, y0, x1, y1);
	}
}

void calc_laser_value ()
{
	// Rotational plane sweep algorithm Ͽ ȿ а  ˻Ѵ.
	// å Principles of Robot Motion, p.114 
	//
	// ⼭ ˰ ϱ  ޸𸮸 Ҵϰ ͸ Ѵ.

	_no_vertex = 2*_map_2d.size();
	_no_edge = _map_2d.size();

	_vertex_list = new sVertex[_no_vertex];
	_edge_list = new sEdge[_no_edge];

	for (unsigned int i=0; i<_map_2d.size(); ++i) {
		_vertex_list[i*2 + 0].x = _map_2d[i].x0;
		_vertex_list[i*2 + 0].y = _map_2d[i].y0;
		_vertex_list[i*2 + 0].alpha = 0.;

		_vertex_list[i*2 + 1].x = _map_2d[i].x1;
		_vertex_list[i*2 + 1].y = _map_2d[i].y1;
		_vertex_list[i*2 + 1].alpha = 0.;

		_edge_list[i].s = &_vertex_list[i*2 + 0];
		_edge_list[i].e = &_vertex_list[i*2 + 1];
	}

	plane_sweep_algorithm ();

	delete [] _vertex_list;
	delete [] _edge_list;
}
